from forepaas.dwh.connect import connect
from forepaas.worker.connector import bulk_insert
import logging
from time import sleep
from requests import get
from datetime import datetime as dt
from datetime import timedelta as td
from os import makedirs
import re
import pandas as pd
import numpy as np
# LZH形式のファイルを解凍するパッケージ lhafile をインポート
import lhafile

# URLの固定部分を指定
FIXED_URL = "http://www1.mbrace.or.jp/od2/K/"

CSV_FILE_HEADER = "レースコード,タイトル,日次,レース日,レース場,\
レース回,天候,風向,風速(m),波の高さ(cm),2連単_組番,3連単_組番,1着_艇番,2着_艇番,3着_艇番,4着_艇番,5着_艇番,6着_艇番\n"

TABLE_NAME = "boat_result_table" # 格納先のテーブル名

logger = logging.getLogger(__name__)

def rename_columns(df): #fileのカラム名とtableのカラム名が違う場合に、変更
    df.rename(columns={
        "タイトル":"title",
        "日次":"daily",
        "レース日":"race_date",
        "レース場":"race_track",
        "レース回":"race_number",
        "3連単_組番":"tierce_number",
        "2連単_組番":"exacta_number",
        "1着_艇番":"place_1",
        "2着_艇番":"place_2",
        "3着_艇番":"place_3",
        "4着_艇番":"place_4",
        "5着_艇番":"place_5",
        "6着_艇番":"place_6",
        "波の高さ(cm)":"wave_height",
        "天候":"weather",
        "風速(m)":"wind_speed",
        "風向":"winds_direction",
        "レースコード":"race_id"
    }, inplace=True)
    return df

# テキストファイルからデータを抽出し、CSVファイルに書き込む関数
def get_data(text_file, path):
    # CSVファイルを追記モードで開く
    csv_file = open(path, "a", encoding="shift_jis")

    # テキストファイルからデータを抽出
    for line in text_file:

        # 文字列「競争成績」を検索してタイトル・日次・レース日・レース場を取得
        # rは正規表現でraw文字列を指定するおまじない
        if re.search(r"競走成績", line):
            # 1行スキップ
            text_file.readline()

            # タイトルを取得
            line = text_file.readline()
            title = line[:-1].strip()

            # 1行スキップ
            text_file.readline()

            # 日次、レース日、レース場を取得
            line = text_file.readline()
            day = line[3:7].replace(' ', '')
            date = line[17:27].replace(' ', '')
            stadium = line[62:65].replace('　', '')
        
        # レース回の「R」と距離の「H」を同じ行に見つけたら -> これ以降に競走成績の詳細が記載
        if re.search(r"R", line) and re.search(r"H", line):

            # レース名にキーワード「進入固定」が割り込んだ際の補正(「進入固定戦隊」は除くためＨまで含めて置換)
            if re.search(r"進入固定", line):
                line = line.replace('進入固定       H', '進入固定           H')

            # レース回、レース名、距離(m)、天候、風向、風速(m)、波の高さ(cm)を取得
            race_round = line[2:5].replace(' ', '0')
            weather = line[43:45].strip()
            wind_direction = line[50:52].strip()
            wind_velocity = line[53:55].strip()
            wave_height = line[60:63].strip()

            # 決まり手を取得
            line = text_file.readline()
            winning_technique = line[50:55].strip()

            # 1行スキップ
            text_file.readline()

            # 選手データを格納する変数を定義
            result_racer = ""

            # 選手データを取り出す行(開始行)を格納
            line = text_file.readline()

            # 不成立の場合回避する用の変数を定義
            result_win = ""
            
            # 空行まで処理を繰り返す = 1～6艇分の選手データを取得
            while line != "\n":
                # 選手データを格納(行末にカンマが入らないように先頭にカンマを入れる)
                result_racer += "," + line[6]

                # 次の行を読み込む
                line = text_file.readline()

            # レース結果を取り出す行(開始行)を格納
            line = text_file.readline()

            # 空行まで処理を繰り返す = レース結果を取得
            while line != "\n":

                # 2連単の結果を取得
                if re.search(r"２連単", line):
                    result_exacta = line[14:17]

                # 3連単の結果を取得
                if re.search(r"３連単", line):
                    result_trifecta = line[14:19]

                # 次の行を読み込む
                line = text_file.readline()

            # レースコードを生成
            dict_stadium = {'桐生': 'KRY', '戸田': 'TDA', '江戸川': 'EDG', '平和島': 'HWJ',
                            '多摩川': 'TMG', '浜名湖': 'HMN', '蒲郡': 'GMG', '常滑': 'TKN',
                            '津': 'TSU', '三国': 'MKN', '琵琶湖': 'BWK', '住之江': 'SME',
                            '尼崎': 'AMG', '鳴門': 'NRT', '丸亀': 'MRG', '児島': 'KJM',
                            '宮島': 'MYJ', '徳山': 'TKY', '下関': 'SMS', '若松': 'WKM',
                            '芦屋': 'ASY', '福岡': 'FKO', '唐津': 'KRT', '大村': 'OMR', 'びわこ': 'BWK'
                            }

            race_code = date.split("/")[0] + date.split("/")[1] + date.split("/")[2] + dict_stadium[stadium] + race_round[0:2]
            # 抽出したデータをCSVファイルに書き込む
            csv_file.write(race_code + "," + title + "," + day + "," + date + "," + stadium \
                        + "," + race_round + "," + weather \
                        + "," + wind_direction + "," + wind_velocity + "," + wave_height \
                        + "," + result_exacta + "," + result_trifecta + result_racer + "\n")

    # CSVファイルを閉じる
    csv_file.close()

def customfunc(event):
    logger.info("Begin function customfunc")

    # URL生成
    today = dt.now().today()
    date = dt.now().today() - td(days=1)
    date_str = date.strftime("%Y%m%d")
    yyyymm = date_str[0:4] + date_str[4:6]
    yymmdd = date_str[2:4] + date_str[4:6] + date_str[6:8]

    variable_url = FIXED_URL + yyyymm + "/k" + yymmdd + ".lzh"
    file_name = "k" + yymmdd + ".lzh"

    # bucket接続
    data_store = connect('data_store')
    bucket_connect = data_store.get_bucket('boat-result-bucket')

    # ダウンロード
    bucket_connect.put_request(variable_url, path=file_name)

    # 解凍
    local_filepath = "/tmp/{}".format(file_name)
    bucket_connect.fget(file_name, local_filepath)  # バケットのfput関数でバケットのファイルをtmp領域に移動します
    unzip_file = lhafile.Lhafile(local_filepath)
    info = unzip_file.infolist()
    name = info[0].filename
    local_unzip_filepath = "/tmp/{}.TXT".format(file_name.split(".")[0])
    open(local_unzip_filepath, "wb").write(unzip_file.read(name))

    # TXTファイルをCSVに変換
    local_csv_path = "/tmp/{}.csv".format(file_name.split(".")[0])
    csv_file = open(local_csv_path, "w", encoding="shift_jis")
    csv_file.write(CSV_FILE_HEADER)
    csv_file.close()

    # テキストファイルを開く
    text_file = open(local_unzip_filepath, "r", encoding="shift_jis")
    # データを抽出する
    get_data(text_file, local_csv_path)
    # テキストファイルを閉じる
    text_file.close()
    # CSVファイルをバケットに保存
    bucket_connect.fput("{}.csv".format(file_name.split(".")[0]), local_csv_path)
    
    # テーブルに追加する
    df = pd.read_csv(local_csv_path, encoding="shift-jis")
    # dfに1個でも欠損値があったら行を削除する
    df.dropna(axis = 0, how = 'any', inplace=True)

    df["1着_艇番"] = df["1着_艇番"].astype(float)
    df["2着_艇番"] = df["2着_艇番"].astype(float)
    df["3着_艇番"] = df["3着_艇番"].astype(float)
    df["4着_艇番"] = df["4着_艇番"].astype(float)
    df["5着_艇番"] = df["5着_艇番"].astype(float)
    df["6着_艇番"] = df["6着_艇番"].astype(float)

    # レースコードを追加
    # レース場コードを生成
    dict_stadium = {'桐生': 'KRY', '戸田': 'TDA', '江戸川': 'EDG', '平和島': 'HWJ',
                    '多摩川': 'TMG', '浜名湖': 'HMN', '蒲郡': 'GMG', '常滑': 'TKN',
                    '津': 'TSU', '三国': 'MKN', '琵琶湖': 'BWK', '住之江': 'SME',
                    '尼崎': 'AMG', '鳴門': 'NRT', '丸亀': 'MRG', '児島': 'KJM',
                    '宮島': 'MYJ', '徳山': 'TKY', '下関': 'SMS', '若松': 'WKM',
                    '芦屋': 'ASY', '福岡': 'FKO', '唐津': 'KRT', '大村': 'OMR','びわこ': 'BWK'
                    }
    df["レースコード"] = df["レース日"].map(lambda x:x.split("/")[0]) + df["レース日"].map(lambda x:x.split("/")[1].zfill(2)) + df["レース日"].map(lambda x:x.split("/")[2].zfill(2)) + df["レース場"].map(lambda x:dict_stadium[x]) +df["レース回"].map(lambda x:x[:-1].zfill(2))
    df = rename_columns(df)
    cn = connect("dwh/data_prim/")
    df.rename(columns = str.lower, inplace=True)
    df_rows = df.shape[0]
    stats = bulk_insert(cn, TABLE_NAME, df)

    # ダウンロードした圧縮ファイル削除
    bucket_connect.delete(file_name)

    # 接続解除
    del data_store
    del bucket_connect
    del cn
    logger.info("END function customfunc")